@@ -1,12 +1,15 @@
 # -*- coding: utf-8 -*-
 
 from attr import attrs, attrib
+import concurrent.futures
 import json
+import queue
 import random
 import re
 import requests
 import time
 import tuxbuild.exceptions
+import uuid
 
 # We will poll for status change for an average duration of 180 minutes
 state_timeout = 10800  # 60 * 180
@@ -59,6 +62,19 @@ def get_request(url, headers):
             r.raise_for_status()  # Some unexpected status that's not caught
 
 
+@attrs(kw_only=True)
+class BuildState:
+    build = attrib()
+    state = attrib()
+    status = attrib(default=None)
+    message = attrib()
+    warnings = attrib(default=0)
+    errors = attrib(default=0)
+    icon = attrib()
+    cli_color = attrib()
+    final = attrib(default=False)
+
+
 @attrs(kw_only=True)
 class Build:
     git_repo = attrib()
@@ -80,6 +96,7 @@ class Build:
             self.kconfig = [self.kconfig]
         self.headers = {"Content-Type": "application/json", "Authorization": self.token}
         self.verify_build_parameters()
+        self.client_token = str(uuid.uuid4())
 
     def __str__(self):
         git_short_log = self.status.get("git_short_log", "")
@@ -89,10 +106,14 @@ class Build:
                 self.kconfig_allconfig
             )
 
+        if len(self.kconfig) > 1:
+            kconfig = f"{self.kconfig[0]}+{len(self.kconfig)-1}"
+        else:
+            kconfig = self.kconfig[0]
         return "{} {} ({}{}) with {} @ {}".format(
             git_short_log,
             self.target_arch,
-            ", ".join(self.kconfig),
+            kconfig,
             kconfig_allconfig_param,
             self.toolchain,
             self.build_data,
@@ -109,6 +130,7 @@ class Build:
     def generate_build_request(self):
         """ return a build data in a python dict """
         build_entry = {}
+        build_entry["client_token"] = self.client_token
         build_entry["git_repo"] = self.git_repo
         if self.git_ref:
             build_entry["git_ref"] = self.git_ref
@@ -170,6 +192,80 @@ class Build:
                 f"Build timed out after {state_timeout/60} minutes; abandoning"
             )
 
+    def watch(self):
+        waiting_states = [
+            BuildState(
+                build=self,
+                state="queued",
+                message="Queued",
+                icon="⏳",
+                cli_color="white",
+            ),
+            BuildState(
+                build=self,
+                state="building",
+                message="Building",
+                icon="⚗️ ",
+                cli_color="cyan",
+            ),
+        ]
+        for state in waiting_states:
+            yield state
+            self.wait_on_status(state.state)
+
+        errors = 0
+        warnings = 0
+        state = self.status["tuxbuild_status"]
+        status = self.status["build_status"]
+
+        if status == "pass":
+            warnings = self.status["warnings_count"]
+            if warnings == 0:
+                icon = "🎉"
+                message = "Pass"
+                cli_color = "green"
+            else:
+                icon = "👾"
+                cli_color = "yellow"
+                if warnings == 1:
+                    message = "Pass (1 warning)"
+                else:
+                    message = "Pass ({} warnings)".format(warnings)
+        elif status == "fail":
+            icon = "👹"
+            cli_color = "bright_red"
+            errors = self.status["errors_count"]
+            if errors == 1:
+                message = "Fail (1 error)"
+            else:
+                message = "Fail ({} errors)".format(errors)
+            error_message = self.status.get("status_message")
+            if error_message:
+                message += f" with status message '{error_message}'"
+        else:
+            icon = "🔧"
+            cli_color = "bright_red"
+            message = self.status["status_message"]
+
+        finished = BuildState(
+            build=self,
+            state=state,
+            status=status,
+            final=True,
+            message=message,
+            icon=icon,
+            cli_color=cli_color,
+            warnings=warnings,
+            errors=errors,
+        )
+        yield finished
+
+    def wait(self):
+        state = None
+        for s in self.watch():
+            state = s
+        return state
+
     def _get_field(self, field):
         """ Retrieve an individual field from status.json """
         self.get_status()
@@ -202,37 +298,60 @@ class Build:
 
 
 class BuildSet:
-    def __init__(self, build_objects, token, kbapi_url):
-        self.build_objects = build_objects
+    def __init__(self, builds, **kwargs):
+        self.builds = []
+        for item in builds:
+            data = kwargs.copy()
+            data.update(item)
+            self.builds.append(Build(**data))
         self.headers = {
             "Content-type": "application/json",
-            "Authorization": "{}".format(token),
+            "Authorization": "{}".format(kwargs["token"]),
         }
-        self.kbapi_url = kbapi_url
+        self.kbapi_url = kwargs["kbapi_url"]
 
     def build(self):
         """ Submit the build request """
 
-        build_objects_list = self.build_objects[:]
         data = []
         url = self.kbapi_url + "/build"
-        for build_object in self.build_objects:
+        for build_object in self.builds:
             data.append(build_object.generate_build_request())
         json_data = post_request(url, self.headers, data)
+
+        builds_by_client_token = {b.client_token: b for b in self.builds}
         for build_config in json_data:
-            for build_object in build_objects_list:
-                if (
-                    build_config.items()
-                    >= build_object.generate_build_request().items()
-                ):
-                    build_object.build_data = build_config["download_url"]
-                    build_object.build_key = build_config["build_key"]
-                    # We may have build set with same parameters multiple times
-                    # hence remove the build from build_objects list after mapping
-                    build_objects_list.remove(build_object)
-                    break
+            build = builds_by_client_token[build_config["client_token"]]
+            build.build_data = build_config["download_url"]
+            build.build_key = build_config["build_key"]
+
+    def __watch_build__(self, build, results):
+        for state in build.watch():
+            results.put(state)
+
+    def watch(self):
+        results = queue.Queue()
+        n = len(self.builds)
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=n) as executor:
+            for build in self.builds:
+                executor.submit(self.__watch_build__, build, results)
+
+            finished = 0
+            while finished < n:
+                state = results.get()
+                yield state
+                if state.final:
+                    finished += 1
+
+    def wait(self):
+        states = []
+        for s in self.watch():
+            if s.final:
+                states.append(s)
+        return states
 
     @property
     def status_list(self):
         """ Return a list of build status dictionaries """
-        return [build.status for build in self.build_objects]
+        return [build.status for build in self.builds]
